"""
Signal constants
"""

SIGNAL_TYPE_V3_VTAB = 0
SIGNAL_TYPE_V3_VTABR = 1
SIGNAL_TYPE_V3_BYTEARRAY = 2
SIGNAL_TYPE_V3_ARRAY = 3

SIGNAL_TYPE_V4_VTAB = 4
SIGNAL_TYPE_V4_VTABR = 5
SIGNAL_TYPE_V4_BYTEARRAY = 6
SIGNAL_TYPE_V4_ARRAY = 7
SIGNAL_TYPE_V4_TTAB = 8
SIGNAL_TYPE_V4_TRANS = 9
SIGNAL_TYPE_V4_CANOPENDATE = 10
SIGNAL_TYPE_V4_CANOPENTIME = 11
